@@ -323,6 +323,10 @@ int Client::handshake_completed() {
     }
     std::cerr << "Negotiated ALPN is " << tls_session_.get_selected_alpn()
               << std::endl;
+
+    if (!config.ech_config_list.empty() && tls_session_.get_ech_accepted()) {
+      std::cerr << "ECH was accepted" << std::endl;
+    }
   }
 
   if (config.tp_file) {
@@ -331,7 +335,7 @@ int Client::handshake_completed() {
       ngtcp2_conn_encode_0rtt_transport_params(conn_, data.data(), data.size());
     if (datalen < 0) {
       std::cerr << "Could not encode 0-RTT transport parameters: "
-                << ngtcp2_strerror(datalen) << std::endl;
+                << ngtcp2_strerror(static_cast<int>(datalen)) << std::endl;
     } else if (util::write_transport_params(
                  config.tp_file, {data.data(), static_cast<size_t>(datalen)}) !=
                0) {
@@ -398,7 +402,8 @@ int recv_version_negotiation(ngtcp2_conn *conn, const ngtcp2_pkt_hd *hd,
 
 void Client::recv_version_negotiation(const uint32_t *sv, size_t nsv) {
   offered_versions_.resize(nsv);
-  std::ranges::copy_n(sv, nsv, std::ranges::begin(offered_versions_));
+  std::ranges::copy_n(sv, as_signed(nsv),
+                      std::ranges::begin(offered_versions_));
 }
 
 namespace {
@@ -665,7 +670,7 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     } else {
       path = std::string{config.qlog_dir};
       path += '/';
-      path += util::format_hex({scid.data, scid.datalen});
+      path += util::format_hex(scid.data, as_signed(scid.datalen));
       path += ".sqlog";
     }
     qlog_ = fopen(path.c_str(), "w");
@@ -696,7 +701,7 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     settings.initial_pkt_num = config.initial_pkt_num;
   }
 
-  std::string token;
+  std::vector<uint8_t> token;
 
   if (!config.token_file.empty()) {
     std::cerr << "Reading token file " << config.token_file << std::endl;
@@ -704,7 +709,7 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
     auto t = util::read_token(config.token_file);
     if (t) {
       token = std::move(*t);
-      settings.token = reinterpret_cast<const uint8_t *>(token.data());
+      settings.token = token.data();
       settings.tokenlen = token.size();
     }
   }
@@ -779,8 +784,7 @@ int Client::init(int fd, const Address &local_addr, const Address &remote_addr,
       early_data_ = false;
     } else {
       auto rv = ngtcp2_conn_decode_and_set_0rtt_transport_params(
-        conn_, reinterpret_cast<const uint8_t *>(params->data()),
-        params->size());
+        conn_, params->data(), params->size());
       if (rv != 0) {
         std::cerr << "ngtcp2_conn_decode_and_set_0rtt_transport_params:"
                   << ngtcp2_strerror(rv) << std::endl;
@@ -820,8 +824,15 @@ int Client::feed_data(const Endpoint &ep, const sockaddr *sa, socklen_t salen,
     std::cerr << "ngtcp2_conn_read_pkt: " << ngtcp2_strerror(rv) << std::endl;
     if (!last_error_.error_code) {
       if (rv == NGTCP2_ERR_CRYPTO) {
-        ngtcp2_ccerr_set_tls_alert(
-          &last_error_, ngtcp2_conn_get_tls_alert(conn_), nullptr, 0);
+        auto alert = ngtcp2_conn_get_tls_alert(conn_);
+        ngtcp2_ccerr_set_tls_alert(&last_error_, alert, nullptr, 0);
+
+        if (alert == TLS_ALERT_ECH_REQUIRED && config.ech_config_list_file &&
+            tls_session_.write_ech_config_list(config.ech_config_list_file) !=
+              0) {
+          std::cerr << "Could not write ECH retry configs in "
+                    << config.ech_config_list_file << std::endl;
+        }
       } else {
         ngtcp2_ccerr_set_liberr(&last_error_, rv, nullptr, 0);
       }
@@ -1031,9 +1042,10 @@ int Client::write_streams() {
 
       assert(ndatalen == -1);
 
-      std::cerr << "ngtcp2_conn_write_stream: " << ngtcp2_strerror(nwrite)
-                << std::endl;
-      ngtcp2_ccerr_set_liberr(&last_error_, nwrite, nullptr, 0);
+      std::cerr << "ngtcp2_conn_write_stream: "
+                << ngtcp2_strerror(static_cast<int>(nwrite)) << std::endl;
+      ngtcp2_ccerr_set_liberr(&last_error_, static_cast<int>(nwrite), nullptr,
+                              0);
       disconnect();
       return -1;
     } else if (ndatalen >= 0) {
@@ -1054,12 +1066,12 @@ int Client::write_streams() {
 
     auto last_pkt_pos = std::ranges::begin(buf);
 
-    buf = buf.subspan(nwrite);
+    buf = buf.subspan(as_unsigned(nwrite));
 
     if (last_pkt_pos == std::ranges::begin(txbuf)) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_ecn = pi.ecn;
-      gso_size = nwrite;
+      gso_size = as_unsigned(nwrite);
     } else if (!ngtcp2_path_eq(&prev_ps.path, &ps.path) || prev_ecn != pi.ecn ||
                static_cast<size_t>(nwrite) > gso_size ||
                (gso_size > path_max_udp_payload_size &&
@@ -1531,12 +1543,18 @@ Client::send_packet(const Endpoint &ep, const ngtcp2_addr &remote_addr,
     cm->cmsg_level = SOL_UDP;
     cm->cmsg_type = UDP_SEGMENT;
     cm->cmsg_len = CMSG_LEN(sizeof(uint16_t));
-    uint16_t n = gso_size;
+    auto n = static_cast<uint16_t>(gso_size);
     memcpy(CMSG_DATA(cm), &n, sizeof(n));
   }
 #endif // defined(UDP_SEGMENT)
 
-  msg.msg_controllen = controllen;
+  msg.msg_controllen =
+#ifndef __APPLE__
+    controllen
+#else  // defined(__APPLE__)
+    static_cast<socklen_t>(controllen)
+#endif // defined(__APPLE__)
+    ;
 
   ssize_t nwrite = 0;
 
@@ -1673,7 +1691,7 @@ int Client::handle_error() {
     util::timestamp());
   if (nwrite < 0) {
     std::cerr << "ngtcp2_conn_write_connection_close: "
-              << ngtcp2_strerror(nwrite) << std::endl;
+              << ngtcp2_strerror(static_cast<int>(nwrite)) << std::endl;
     return -1;
   }
 
@@ -2196,6 +2214,12 @@ Options:
   --pmtud-probes=<SIZE>[[,<SIZE>]...]
               Specify UDP datagram payload sizes  to probe in Path MTU
               Discovery.  <SIZE> must be strictly larger than 1200.
+  --ech-config-list-file=<PATH>
+              Read/write  ECHConfigList from/to  <PATH>.  ECH  is only
+              attempted if  an underlying  TLS stack supports  it.  If
+              the handshake  fails with ech_required alert,  ECH retry
+              configs,  if  provided by  server,  will  be written  to
+              <PATH>.
   -h, --help  Display this help and exit.
 
 ---
@@ -2279,6 +2303,7 @@ int main(int argc, char **argv) {
       {"wait-for-ticket", no_argument, &flag, 41},
       {"initial-pkt-num", required_argument, &flag, 42},
       {"pmtud-probes", required_argument, &flag, 43},
+      {"ech-config-list-file", required_argument, &flag, 44},
       {},
     };
 
@@ -2708,11 +2733,15 @@ int main(int argc, char **argv) {
               << std::endl;
             exit(EXIT_FAILURE);
           } else {
-            config.pmtud_probes.push_back(*n);
+            config.pmtud_probes.push_back(static_cast<uint16_t>(*n));
           }
         }
         break;
       }
+      case 44:
+        // --ech-config-list-file
+        config.ech_config_list_file = optarg;
+        break;
       }
       break;
     default:
@@ -2757,7 +2786,7 @@ int main(int argc, char **argv) {
       exit(EXIT_FAILURE);
     }
     config.fd = fd;
-    config.datalen = st.st_size;
+    config.datalen = static_cast<size_t>(st.st_size);
     if (config.datalen) {
       auto addr = mmap(nullptr, config.datalen, PROT_READ, MAP_SHARED, fd, 0);
       if (addr == MAP_FAILED) {
@@ -2769,10 +2798,20 @@ int main(int argc, char **argv) {
     }
   }
 
+  if (config.ech_config_list_file) {
+    auto ech_config = util::read_file(config.ech_config_list_file);
+    if (!ech_config) {
+      std::cerr << "ech-config-list-file: Could not read ECHConfigList"
+                << std::endl;
+    } else {
+      config.ech_config_list = std::move(*ech_config);
+    }
+  }
+
   auto addr = argv[optind++];
   auto port = argv[optind++];
 
-  if (parse_requests(&argv[optind], argc - optind) != 0) {
+  if (parse_requests(&argv[optind], static_cast<size_t>(argc - optind)) != 0) {
     exit(EXIT_FAILURE);
   }
 
