@@ -130,7 +130,7 @@ Request request_path(const std::string_view &uri) {
     req.path = std::string(uri.data() + u.field_data[URLPARSE_PATH].off,
                            u.field_data[URLPARSE_PATH].len);
     if (req.path.find('%') != std::string::npos) {
-      req.path = util::percent_decode(std::begin(req.path), std::end(req.path));
+      req.path = util::percent_decode(req.path);
     }
     if (!req.path.empty() && req.path.back() == '/') {
       req.path += "index.html";
@@ -165,7 +165,7 @@ std::unordered_map<std::string, FileEntry> file_cache;
 
 std::pair<FileEntry, int> Stream::open_file(const std::string &path) {
   auto it = file_cache.find(path);
-  if (it != std::end(file_cache)) {
+  if (it != std::ranges::end(file_cache)) {
     return {(*it).second, 0};
   }
 
@@ -187,7 +187,7 @@ std::pair<FileEntry, int> Stream::open_file(const std::string &path) {
     close(fd);
   } else {
     fe.fd = fd;
-    fe.len = st.st_size;
+    fe.len = static_cast<size_t>(st.st_size);
     if (fe.len) {
       fe.map = mmap(nullptr, fe.len, PROT_READ, MAP_SHARED, fd, 0);
       if (fe.map == MAP_FAILED) {
@@ -407,9 +407,7 @@ int Handler::handshake_completed() {
   std::array<uint8_t, NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN> token;
 
   auto path = ngtcp2_conn_get_path(conn_);
-  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
-             std::chrono::system_clock::now().time_since_epoch())
-             .count();
+  auto t = util::system_clock_now();
 
   auto tokenlen = ngtcp2_crypto_generate_regular_token(
     token.data(), config.static_secret.data(), config.static_secret.size(),
@@ -421,7 +419,8 @@ int Handler::handshake_completed() {
     return 0;
   }
 
-  if (auto rv = ngtcp2_conn_submit_new_token(conn_, token.data(), tokenlen);
+  if (auto rv = ngtcp2_conn_submit_new_token(conn_, token.data(),
+                                             as_unsigned(tokenlen));
       rv != 0) {
     if (!config.quiet) {
       std::cerr << "ngtcp2_conn_submit_new_token: " << ngtcp2_strerror(rv)
@@ -491,7 +490,7 @@ int acked_stream_data_offset(ngtcp2_conn *conn, int64_t stream_id,
 int Handler::acked_stream_data_offset(int64_t stream_id, uint64_t offset,
                                       uint64_t datalen) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
   (void)stream;
 
@@ -515,7 +514,7 @@ void Handler::on_stream_open(int64_t stream_id) {
   }
   auto it = streams_.find(stream_id);
   (void)it;
-  assert(it == std::end(streams_));
+  assert(it == std::ranges::end(streams_));
   streams_.emplace(stream_id, std::make_unique<Stream>(stream_id, this));
 }
 
@@ -602,9 +601,7 @@ int path_validation(ngtcp2_conn *conn, uint32_t flags, const ngtcp2_path *path,
   }
 
   std::array<uint8_t, NGTCP2_CRYPTO_MAX_REGULAR_TOKENLEN> token;
-  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
-             std::chrono::system_clock::now().time_since_epoch())
-             .count();
+  auto t = util::system_clock_now();
 
   auto tokenlen = ngtcp2_crypto_generate_regular_token(
     token.data(), config.static_secret.data(), config.static_secret.size(),
@@ -617,7 +614,8 @@ int path_validation(ngtcp2_conn *conn, uint32_t flags, const ngtcp2_path *path,
     return 0;
   }
 
-  if (auto rv = ngtcp2_conn_submit_new_token(conn, token.data(), tokenlen);
+  if (auto rv =
+        ngtcp2_conn_submit_new_token(conn, token.data(), as_unsigned(tokenlen));
       rv != 0) {
     if (!config.quiet) {
       std::cerr << "ngtcp2_conn_submit_new_token: " << ngtcp2_strerror(rv)
@@ -645,7 +643,7 @@ int extend_max_stream_data(ngtcp2_conn *conn, int64_t stream_id,
 
 int Handler::extend_max_stream_data(int64_t stream_id, uint64_t max_data) {
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   if (nghttp3_buf_len(&stream->respbuf)) {
@@ -723,7 +721,7 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
   if (!config.qlog_dir.empty()) {
     auto path = std::string{config.qlog_dir};
     path += '/';
-    path += util::format_hex({scid_.data, scid_.datalen});
+    path += util::format_hex(scid_.data, as_signed(scid_.datalen));
     path += ".sqlog";
     qlog_ = fopen(path.c_str(), "w");
     if (qlog_ == nullptr) {
@@ -753,8 +751,8 @@ int Handler::init(const Endpoint &ep, const Address &local_addr,
     settings.pmtud_probeslen = config.pmtud_probes.size();
 
     if (!config.max_udp_payload_size) {
-      settings.max_tx_udp_payload_size = *std::max_element(
-        std::begin(config.pmtud_probes), std::end(config.pmtud_probes));
+      settings.max_tx_udp_payload_size =
+        *std::ranges::max_element(config.pmtud_probes);
     }
   }
 
@@ -964,6 +962,8 @@ int Handler::write_streams() {
     std::span{tx_.data.get(), std::max(ngtcp2_conn_get_send_quantum(conn_),
                                        path_max_udp_payload_size)};
   auto buf = txbuf;
+  auto pkt = std::span<const uint8_t>{};
+  auto extra_pkt = std::span<const uint8_t>{};
 
   ngtcp2_path_storage_zero(&ps);
   ngtcp2_path_storage_zero(&prev_ps);
@@ -976,7 +976,7 @@ int Handler::write_streams() {
     Stream *stream = nullptr;
 
     if (!sendq_.empty() && ngtcp2_conn_get_max_data_left(conn_)) {
-      stream = *std::begin(sendq_);
+      stream = *std::ranges::begin(sendq_);
 
       stream_id = stream->stream_id;
       vec.base = stream->respbuf.pos;
@@ -998,129 +998,98 @@ int Handler::write_streams() {
       case NGTCP2_ERR_STREAM_DATA_BLOCKED:
       case NGTCP2_ERR_STREAM_SHUT_WR:
         assert(ndatalen == -1);
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
         continue;
       case NGTCP2_ERR_WRITE_MORE:
         assert(ndatalen >= 0);
         stream->respbuf.pos += ndatalen;
         if (nghttp3_buf_len(&stream->respbuf) == 0) {
-          sendq_.erase(std::begin(sendq_));
+          sendq_.erase(std::ranges::begin(sendq_));
         }
         continue;
       }
 
       assert(ndatalen == -1);
 
-      std::cerr << "ngtcp2_conn_writev_stream: " << ngtcp2_strerror(nwrite)
-                << std::endl;
-      ngtcp2_ccerr_set_liberr(&last_error_, nwrite, nullptr, 0);
+      std::cerr << "ngtcp2_conn_writev_stream: "
+                << ngtcp2_strerror(static_cast<int>(nwrite)) << std::endl;
+      ngtcp2_ccerr_set_liberr(&last_error_, static_cast<int>(nwrite), nullptr,
+                              0);
       return handle_error();
     } else if (ndatalen >= 0) {
       stream->respbuf.pos += ndatalen;
       if (nghttp3_buf_len(&stream->respbuf) == 0) {
-        sendq_.erase(std::begin(sendq_));
+        sendq_.erase(std::ranges::begin(sendq_));
       }
     }
 
     if (nwrite == 0) {
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-      if (!data.empty()) {
-        auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-
-        if (auto [rest, rv] = server_->send_packet(
-              ep, no_gso_, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-              data, gso_size);
-            rv != NETWORK_ERR_OK) {
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                          rest, gso_size);
-
-          start_wev_endpoint(ep);
-        }
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      if (pkt.empty()) {
+        return 0;
       }
 
-      // We are congestion limited.
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      break;
     }
 
-    auto last_pkt = std::begin(buf);
+    auto last_pkt_pos = std::ranges::begin(buf);
 
-    buf = buf.subspan(nwrite);
+    buf = buf.subspan(as_unsigned(nwrite));
 
-    if (last_pkt == std::begin(txbuf)) {
+    if (last_pkt_pos == std::ranges::begin(txbuf)) {
       ngtcp2_path_copy(&prev_ps.path, &ps.path);
       prev_ecn = pi.ecn;
-      gso_size = nwrite;
+      gso_size = as_unsigned(nwrite);
     } else if (!ngtcp2_path_eq(&prev_ps.path, &ps.path) || prev_ecn != pi.ecn ||
                static_cast<size_t>(nwrite) > gso_size ||
                (gso_size > path_max_udp_payload_size &&
                 static_cast<size_t>(nwrite) != gso_size)) {
-      auto &ep = *static_cast<Endpoint *>(prev_ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), last_pkt};
-
-      if (auto [rest, rv] =
-            server_->send_packet(ep, no_gso_, prev_ps.path.local,
-                                 prev_ps.path.remote, prev_ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, prev_ps.path.local, prev_ps.path.remote, prev_ecn,
-                        rest, gso_size);
-
-        data = std::span{last_pkt, std::begin(buf)};
-        on_send_blocked(*static_cast<Endpoint *>(ps.path.user_data),
-                        ps.path.local, ps.path.remote, pi.ecn, data,
-                        data.size());
-
-        start_wev_endpoint(ep);
-      } else {
-        auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-        auto data = std::span{last_pkt, std::begin(buf)};
-
-        if (auto [rest, rv] =
-              server_->send_packet(ep, no_gso_, ps.path.local, ps.path.remote,
-                                   pi.ecn, data, data.size());
-            rv != 0) {
-          assert(rest.size() == data.size());
-          assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-          on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
-                          rest.size());
-
-          start_wev_endpoint(ep);
-        }
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), last_pkt_pos};
+      extra_pkt = std::span{last_pkt_pos, std::ranges::begin(buf)};
+      break;
     }
 
     if (buf.size() < path_max_udp_payload_size ||
         static_cast<size_t>(nwrite) < gso_size) {
-      auto &ep = *static_cast<Endpoint *>(ps.path.user_data);
-      auto data = std::span{std::begin(txbuf), std::begin(buf)};
-
-      if (auto [rest, rv] = server_->send_packet(
-            ep, no_gso_, ps.path.local, ps.path.remote, pi.ecn, data, gso_size);
-          rv != 0) {
-        assert(NETWORK_ERR_SEND_BLOCKED == rv);
-
-        on_send_blocked(ep, ps.path.local, ps.path.remote, pi.ecn, rest,
-                        gso_size);
-
-        start_wev_endpoint(ep);
-      }
-
-      ngtcp2_conn_update_pkt_tx_time(conn_, ts);
-      return 0;
+      pkt = std::span{std::ranges::begin(txbuf), std::ranges::begin(buf)};
+      break;
     }
   }
+
+  assert(!pkt.empty());
+
+  if (send_packet(prev_ps.path, prev_ecn, pkt, gso_size) != 0) {
+    if (!extra_pkt.empty()) {
+      on_send_blocked(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+    }
+  } else if (!extra_pkt.empty()) {
+    send_packet(ps.path, pi.ecn, extra_pkt, extra_pkt.size());
+  }
+
+  ngtcp2_conn_update_pkt_tx_time(conn_, ts);
+
+  return 0;
 }
 
-void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
-                              const ngtcp2_addr &remote_addr, unsigned int ecn,
+int Handler::send_packet(const ngtcp2_path &path, unsigned int ecn,
+                         std::span<const uint8_t> data, size_t gso_size) {
+  auto &ep = *static_cast<Endpoint *>(path.user_data);
+  auto [rest, rv] = server_->send_packet(ep, no_gso_, path.local, path.remote,
+                                         ecn, data, gso_size);
+  if (rv != 0) {
+    assert(NETWORK_ERR_SEND_BLOCKED == rv);
+
+    on_send_blocked(path, ecn, rest, gso_size);
+
+    start_wev_endpoint(ep);
+
+    return rv;
+  }
+
+  return 0;
+}
+
+void Handler::on_send_blocked(const ngtcp2_path &path, unsigned int ecn,
                               std::span<const uint8_t> data, size_t gso_size) {
   assert(tx_.num_blocked || !tx_.send_blocked);
   assert(tx_.num_blocked < 2);
@@ -1130,12 +1099,12 @@ void Handler::on_send_blocked(Endpoint &ep, const ngtcp2_addr &local_addr,
 
   auto &p = tx_.blocked[tx_.num_blocked++];
 
-  memcpy(&p.local_addr.su, local_addr.addr, local_addr.addrlen);
-  memcpy(&p.remote_addr.su, remote_addr.addr, remote_addr.addrlen);
+  memcpy(&p.local_addr.su, path.local.addr, path.local.addrlen);
+  memcpy(&p.remote_addr.su, path.remote.addr, path.remote.addrlen);
 
-  p.local_addr.len = local_addr.addrlen;
-  p.remote_addr.len = remote_addr.addrlen;
-  p.endpoint = &ep;
+  p.local_addr.len = path.local.addrlen;
+  p.remote_addr.len = path.remote.addrlen;
+  p.endpoint = static_cast<Endpoint *>(path.user_data);
   p.ecn = ecn;
   p.data = data;
   p.gso_size = gso_size;
@@ -1235,8 +1204,8 @@ int Handler::start_closing_period() {
     conn_, &ps.path, &pi, conn_closebuf_->wpos(), conn_closebuf_->left(),
     &last_error_, util::timestamp());
   if (n < 0) {
-    std::cerr << "ngtcp2_conn_write_connection_close: " << ngtcp2_strerror(n)
-              << std::endl;
+    std::cerr << "ngtcp2_conn_write_connection_close: "
+              << ngtcp2_strerror(static_cast<int>(n)) << std::endl;
     return -1;
   }
 
@@ -1244,7 +1213,7 @@ int Handler::start_closing_period() {
     return 0;
   }
 
-  conn_closebuf_->push(n);
+  conn_closebuf_->push(as_unsigned(n));
 
   return 0;
 }
@@ -1393,7 +1362,7 @@ int Handler::recv_stream_data(uint32_t flags, int64_t stream_id,
   }
 
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   if (!stream->eos) {
@@ -1459,7 +1428,7 @@ int Handler::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
   }
 
   auto it = streams_.find(stream_id);
-  assert(it != std::end(streams_));
+  assert(it != std::ranges::end(streams_));
   auto &stream = (*it).second;
 
   sendq_.erase(stream.get());
@@ -1479,7 +1448,7 @@ int Handler::on_stream_close(int64_t stream_id, uint64_t app_error_code) {
   return 0;
 }
 
-void Handler::shutdown_read(int64_t stream_id, int app_error_code) {
+void Handler::shutdown_read(int64_t stream_id, uint64_t app_error_code) {
   ngtcp2_conn_shutdown_stream_read(conn_, 0, stream_id, app_error_code);
 }
 
@@ -1532,7 +1501,7 @@ void Server::disconnect() {
   ev_signal_stop(loop_, &sigintev_);
 
   while (!handlers_.empty()) {
-    auto it = std::begin(handlers_);
+    auto it = std::ranges::begin(handlers_);
     auto &h = (*it).second;
 
     h->handle_error();
@@ -1607,6 +1576,7 @@ int create_sock(Address &local_addr, const char *addr, const char *port,
     fd_set_recv_ecn(fd, rp->ai_family);
     fd_set_ip_mtu_discover(fd, rp->ai_family);
     fd_set_ip_dontfrag(fd, family);
+    fd_set_udp_gro(fd);
 
     if (bind(fd, rp->ai_addr, rp->ai_addrlen) != -1) {
       break;
@@ -1692,6 +1662,7 @@ int add_endpoint(std::vector<Endpoint> &endpoints, const Address &addr) {
   fd_set_recv_ecn(fd, addr.su.sa.sa_family);
   fd_set_ip_mtu_discover(fd, addr.su.sa.sa_family);
   fd_set_ip_dontfrag(fd, addr.su.sa.sa_family);
+  fd_set_udp_gro(fd);
 
   if (bind(fd, &addr.su.sa, addr.len) == -1) {
     std::cerr << "bind: " << strerror(errno) << std::endl;
@@ -1748,7 +1719,7 @@ int Server::init(const char *addr, const char *port) {
   return 0;
 }
 
-int Server::on_read(Endpoint &ep) {
+int Server::on_read(const Endpoint &ep) {
   sockaddr_union su;
   std::array<uint8_t, 64_k> buf;
   size_t pktcnt = 0;
@@ -1848,7 +1819,7 @@ int Server::on_read(Endpoint &ep) {
   return 0;
 }
 
-void Server::read_pkt(Endpoint &ep, const Address &local_addr,
+void Server::read_pkt(const Endpoint &ep, const Address &local_addr,
                       const sockaddr *sa, socklen_t salen,
                       const ngtcp2_pkt_info *pi,
                       std::span<const uint8_t> data) {
@@ -1872,7 +1843,7 @@ void Server::read_pkt(Endpoint &ep, const Address &local_addr,
   auto dcid_key = util::make_cid_key({vc.dcid, vc.dcidlen});
 
   auto handler_it = handlers_.find(dcid_key);
-  if (handler_it == std::end(handlers_)) {
+  if (handler_it == std::ranges::end(handlers_)) {
     ngtcp2_pkt_hd hd;
 
     if (auto rv = ngtcp2_accept(&hd, data.data(), data.size()); rv != 0) {
@@ -2040,12 +2011,13 @@ uint32_t generate_reserved_version(const sockaddr *sa, socklen_t salen,
 int Server::send_version_negotiation(uint32_t version,
                                      std::span<const uint8_t> dcid,
                                      std::span<const uint8_t> scid,
-                                     Endpoint &ep, const Address &local_addr,
+                                     const Endpoint &ep,
+                                     const Address &local_addr,
                                      const sockaddr *sa, socklen_t salen) {
   Buffer buf{NGTCP2_MAX_UDP_PAYLOAD_SIZE};
   std::array<uint32_t, 1 + max_preferred_versionslen> sv;
 
-  auto p = std::begin(sv);
+  auto p = std::ranges::begin(sv);
 
   *p++ = generate_reserved_version(sa, salen, version);
 
@@ -2060,14 +2032,14 @@ int Server::send_version_negotiation(uint32_t version,
   auto nwrite = ngtcp2_pkt_write_version_negotiation(
     buf.wpos(), buf.left(), std::uniform_int_distribution<uint8_t>()(randgen),
     dcid.data(), dcid.size(), scid.data(), scid.size(), sv.data(),
-    p - std::begin(sv));
+    as_unsigned(p - std::ranges::begin(sv)));
   if (nwrite < 0) {
     std::cerr << "ngtcp2_pkt_write_version_negotiation: "
-              << ngtcp2_strerror(nwrite) << std::endl;
+              << ngtcp2_strerror(static_cast<int>(nwrite)) << std::endl;
     return -1;
   }
 
-  buf.push(nwrite);
+  buf.push(as_unsigned(nwrite));
 
   ngtcp2_addr laddr{
     .addr = const_cast<sockaddr *>(&local_addr.su.sa),
@@ -2086,7 +2058,7 @@ int Server::send_version_negotiation(uint32_t version,
   return 0;
 }
 
-int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
+int Server::send_retry(const ngtcp2_pkt_hd *chd, const Endpoint &ep,
                        const Address &local_addr, const sockaddr *sa,
                        socklen_t salen, size_t max_pktlen) {
   std::array<char, NI_MAXHOST> host;
@@ -2113,9 +2085,7 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
 
   std::array<uint8_t, NGTCP2_CRYPTO_MAX_RETRY_TOKENLEN2> token;
 
-  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
-             std::chrono::system_clock::now().time_since_epoch())
-             .count();
+  auto t = util::system_clock_now();
 
   auto tokenlen = ngtcp2_crypto_generate_retry_token2(
     token.data(), config.static_secret.data(), config.static_secret.size(),
@@ -2126,21 +2096,21 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
 
   if (!config.quiet) {
     std::cerr << "Generated address validation token:" << std::endl;
-    util::hexdump(stderr, token.data(), tokenlen);
+    util::hexdump(stderr, token.data(), as_unsigned(tokenlen));
   }
 
   Buffer buf{
     std::min(static_cast<size_t>(NGTCP2_MAX_UDP_PAYLOAD_SIZE), max_pktlen)};
 
-  auto nwrite =
-    ngtcp2_crypto_write_retry(buf.wpos(), buf.left(), chd->version, &chd->scid,
-                              &scid, &chd->dcid, token.data(), tokenlen);
+  auto nwrite = ngtcp2_crypto_write_retry(buf.wpos(), buf.left(), chd->version,
+                                          &chd->scid, &scid, &chd->dcid,
+                                          token.data(), as_unsigned(tokenlen));
   if (nwrite < 0) {
     std::cerr << "ngtcp2_crypto_write_retry failed" << std::endl;
     return -1;
   }
 
-  buf.push(nwrite);
+  buf.push(as_unsigned(nwrite));
 
   ngtcp2_addr laddr{
     .addr = const_cast<sockaddr *>(&local_addr.su.sa),
@@ -2160,7 +2130,7 @@ int Server::send_retry(const ngtcp2_pkt_hd *chd, Endpoint &ep,
 }
 
 int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
-                                            Endpoint &ep,
+                                            const Endpoint &ep,
                                             const Address &local_addr,
                                             const sockaddr *sa,
                                             socklen_t salen) {
@@ -2174,7 +2144,7 @@ int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
     return -1;
   }
 
-  buf.push(nwrite);
+  buf.push(as_unsigned(nwrite));
 
   ngtcp2_addr laddr{
     .addr = const_cast<sockaddr *>(&local_addr.su.sa),
@@ -2194,7 +2164,7 @@ int Server::send_stateless_connection_close(const ngtcp2_pkt_hd *chd,
 }
 
 int Server::send_stateless_reset(size_t pktlen, std::span<const uint8_t> dcid,
-                                 Endpoint &ep, const Address &local_addr,
+                                 const Endpoint &ep, const Address &local_addr,
                                  const sockaddr *sa, socklen_t salen) {
   if (stateless_reset_bucket_ == 0) {
     return 0;
@@ -2243,13 +2213,13 @@ int Server::send_stateless_reset(size_t pktlen, std::span<const uint8_t> dcid,
   auto nwrite = ngtcp2_pkt_write_stateless_reset(
     buf.wpos(), buf.left(), token.data(), rand_bytes.data(), rand_byteslen);
   if (nwrite < 0) {
-    std::cerr << "ngtcp2_pkt_write_stateless_reset: " << ngtcp2_strerror(nwrite)
-              << std::endl;
+    std::cerr << "ngtcp2_pkt_write_stateless_reset: "
+              << ngtcp2_strerror(static_cast<int>(nwrite)) << std::endl;
 
     return -1;
   }
 
-  buf.push(nwrite);
+  buf.push(as_unsigned(nwrite));
 
   ngtcp2_addr laddr{
     .addr = const_cast<sockaddr *>(&local_addr.su.sa),
@@ -2288,9 +2258,7 @@ int Server::verify_retry_token(ngtcp2_cid *ocid, const ngtcp2_pkt_hd *hd,
     util::hexdump(stderr, hd->token, hd->tokenlen);
   }
 
-  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
-             std::chrono::system_clock::now().time_since_epoch())
-             .count();
+  auto t = util::system_clock_now();
 
   rv = ngtcp2_crypto_verify_retry_token2(
     ocid, hd->token, hd->tokenlen, config.static_secret.data(),
@@ -2335,9 +2303,7 @@ int Server::verify_token(const ngtcp2_pkt_hd *hd, const sockaddr *sa,
     util::hexdump(stderr, hd->token, hd->tokenlen);
   }
 
-  auto t = std::chrono::duration_cast<std::chrono::nanoseconds>(
-             std::chrono::system_clock::now().time_since_epoch())
-             .count();
+  auto t = util::system_clock_now();
 
   if (ngtcp2_crypto_verify_regular_token(hd->token, hd->tokenlen,
                                          config.static_secret.data(),
@@ -2384,7 +2350,7 @@ Server::send_packet(const Endpoint &ep, bool &no_gso,
       auto len = std::min(gso_size, data.size());
 
       auto [_, rv] = send_packet(ep, no_gso, local_addr, remote_addr, ecn,
-                                 {std::begin(data), len}, len);
+                                 data.first(len), len);
       if (rv != 0) {
         return {data, rv};
       }
@@ -2454,7 +2420,7 @@ Server::send_packet(const Endpoint &ep, bool &no_gso,
     cm->cmsg_level = SOL_UDP;
     cm->cmsg_type = UDP_SEGMENT;
     cm->cmsg_len = CMSG_LEN(sizeof(uint16_t));
-    uint16_t n = gso_size;
+    auto n = static_cast<uint16_t>(gso_size);
     memcpy(CMSG_DATA(cm), &n, sizeof(n));
   }
 #endif // defined(UDP_SEGMENT)
@@ -2479,7 +2445,13 @@ Server::send_packet(const Endpoint &ep, bool &no_gso,
     assert(0);
   }
 
-  msg.msg_controllen = controllen;
+  msg.msg_controllen =
+#ifndef __APPLE__
+    controllen
+#else  // defined(__APPLE__)
+    static_cast<socklen_t>(controllen)
+#endif // defined(__APPLE__)
+    ;
 
   ssize_t nwrite = 0;
 
@@ -2558,49 +2530,56 @@ void Server::on_stateless_reset_regen() {
 }
 
 namespace {
-int parse_host_port(Address &dest, int af, const char *first,
-                    const char *last) {
-  if (std::distance(first, last) == 0) {
+int parse_host_port(Address &dest, int af, const std::string_view &host_port) {
+  if (host_port.empty()) {
     return -1;
   }
 
-  const char *host_begin, *host_end, *it;
+  auto first = std::ranges::begin(host_port);
+  auto last = std::ranges::end(host_port);
+
+  std::string_view hostv;
+
   if (*first == '[') {
-    host_begin = first + 1;
-    it = std::find(host_begin, last, ']');
+    ++first;
+
+    auto it = std::ranges::find(first, last, ']');
     if (it == last) {
       return -1;
     }
-    host_end = it;
-    ++it;
-    if (it == last || *it != ':') {
+
+    hostv = std::string_view{first, it};
+    first = it + 1;
+
+    if (first == last || *first != ':') {
       return -1;
     }
   } else {
-    host_begin = first;
-    it = std::find(host_begin, last, ':');
+    auto it = std::ranges::find(first, last, ':');
     if (it == last) {
       return -1;
     }
-    host_end = it;
+
+    hostv = std::string_view{first, it};
+    first = it;
   }
 
-  if (++it == last) {
+  if (++first == last) {
     return -1;
   }
-  auto svc_begin = it;
 
   std::array<char, NI_MAXHOST> host;
-  *std::copy(host_begin, host_end, std::begin(host)) = '\0';
+  *std::ranges::copy(hostv, std::ranges::begin(host)).out = '\0';
 
   addrinfo hints{
     .ai_family = af,
     .ai_socktype = SOCK_DGRAM,
   };
   addrinfo *res;
+  auto svc = first;
 
-  if (auto rv = getaddrinfo(host.data(), svc_begin, &hints, &res); rv != 0) {
-    std::cerr << "getaddrinfo: [" << host.data() << "]:" << svc_begin << ": "
+  if (auto rv = getaddrinfo(host.data(), svc, &hints, &res); rv != 0) {
+    std::cerr << "getaddrinfo: [" << host.data() << "]:" << svc << ": "
               << gai_strerror(rv) << std::endl;
     return -1;
   }
@@ -2821,6 +2800,13 @@ Options:
   --pmtud-probes=<SIZE>[[,<SIZE>]...]
               Specify UDP datagram payload sizes  to probe in Path MTU
               Discovery.  <SIZE> must be strictly larger than 1200.
+  --ech-config-file=<PATH>
+              Read private  key and  ECHConfig from |PATH|.   The file
+              denoted by |PATH| must contain private key and ECHConfig
+              as                      described                     in
+              https://datatracker.ietf.org/doc/html/draft-farrell-tls-pemesni.
+              ECH configuration  is only applied if  an underlying TLS
+              stack supports it.
   -h, --help  Display this help and exit.
 
 ---
@@ -2849,6 +2835,8 @@ int main(int argc, char **argv) {
     prog = basename(argv[0]);
   }
 
+  std::string_view ech_config_file;
+
   for (;;) {
     static int flag = 0;
     constexpr static option long_opts[] = {
@@ -2890,6 +2878,7 @@ int main(int argc, char **argv) {
       {"ack-thresh", required_argument, &flag, 30},
       {"initial-pkt-num", required_argument, &flag, 31},
       {"pmtud-probes", required_argument, &flag, 32},
+      {"ech-config-file", required_argument, &flag, 33},
       {},
     };
 
@@ -2962,8 +2951,7 @@ int main(int argc, char **argv) {
         break;
       case 4:
         // --preferred-ipv4-addr
-        if (parse_host_port(config.preferred_ipv4_addr, AF_INET, optarg,
-                            optarg + strlen(optarg)) != 0) {
+        if (parse_host_port(config.preferred_ipv4_addr, AF_INET, optarg) != 0) {
           std::cerr << "preferred-ipv4-addr: could not use "
                     << std::quoted(optarg) << std::endl;
           exit(EXIT_FAILURE);
@@ -2971,8 +2959,8 @@ int main(int argc, char **argv) {
         break;
       case 5:
         // --preferred-ipv6-addr
-        if (parse_host_port(config.preferred_ipv6_addr, AF_INET6, optarg,
-                            optarg + strlen(optarg)) != 0) {
+        if (parse_host_port(config.preferred_ipv6_addr, AF_INET6, optarg) !=
+            0) {
           std::cerr << "preferred-ipv6-addr: could not use "
                     << std::quoted(optarg) << std::endl;
           exit(EXIT_FAILURE);
@@ -3144,7 +3132,7 @@ int main(int argc, char **argv) {
                     << max_preferred_versionslen << std::endl;
         }
         config.preferred_versions.resize(l.size());
-        auto it = std::begin(config.preferred_versions);
+        auto it = std::ranges::begin(config.preferred_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
@@ -3173,7 +3161,7 @@ int main(int argc, char **argv) {
         // --available-versions
         auto l = util::split_str(optarg);
         config.available_versions.resize(l.size());
-        auto it = std::begin(config.available_versions);
+        auto it = std::ranges::begin(config.available_versions);
         for (const auto &k : l) {
           if (k == "v1"sv) {
             *it++ = NGTCP2_PROTO_VER_V1;
@@ -3235,11 +3223,15 @@ int main(int argc, char **argv) {
               << std::endl;
             exit(EXIT_FAILURE);
           } else {
-            config.pmtud_probes.push_back(*n);
+            config.pmtud_probes.push_back(static_cast<uint16_t>(*n));
           }
         }
         break;
       }
+      case 33:
+        // --ech-config-file
+        ech_config_file = optarg;
+        break;
       }
       break;
     default:
@@ -3265,7 +3257,7 @@ int main(int argc, char **argv) {
     std::cerr << "port: must not exceed 65535" << std::endl;
     exit(EXIT_FAILURE);
   } else {
-    config.port = *n;
+    config.port = static_cast<uint16_t>(*n);
   }
 
   if (auto mt = util::read_mime_types(config.mime_types_file); !mt) {
@@ -3275,6 +3267,17 @@ int main(int argc, char **argv) {
     config.mime_types = std::move(*mt);
   }
 
+  if (!ech_config_file.empty()) {
+    auto ech_config = util::read_ech_server_config(ech_config_file);
+    if (!ech_config) {
+      std::cerr << "ech-config-file: Could not read private key and ECHConfig"
+                << std::endl;
+      exit(EXIT_FAILURE);
+    }
+
+    config.ech_config = std::move(*ech_config);
+  }
+
   TLSServerContext tls_ctx;
 
   if (tls_ctx.init(private_key_file, cert_file, AppProtocol::HQ) != 0) {
